#  Drakkar-Software OctoBot-Node
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

import pytest

from pytest_mock import MockerFixture

from octobot_node.scheduler.task_context import encrypted_task
from octobot_node.app.models import Task
from octobot_node.app.enums import TaskResultKeys


class TestEncryptedTask:
    def test_encrypted_task_no_encryption_keys(
        self, mocker: MockerFixture
    ) -> None:
        mock_settings = mocker.patch("octobot_node.scheduler.task_context.settings")
        mock_settings.TASKS_INPUTS_RSA_PRIVATE_KEY = None
        mock_settings.TASKS_INPUTS_ECDSA_PUBLIC_KEY = None
        mock_settings.TASKS_OUTPUTS_RSA_PUBLIC_KEY = None
        mock_settings.TASKS_OUTPUTS_ECDSA_PRIVATE_KEY = None

        task = Task(
            name="test_task",
            content="plain content",
            result=None
        )
        original_content = task.content

        with encrypted_task(task):
            # Content should remain unchanged
            assert task.content == original_content
            # Can modify task inside context
            task.result = {"status": "success"}

        # Content should still be unchanged after context
        assert task.content == original_content
        assert task.result == {"status": "success"}

    def test_encrypted_task_full_encryption_decryption(
        self, mocker: MockerFixture
    ) -> None:
        mock_settings = mocker.patch("octobot_node.scheduler.task_context.settings")
        mock_settings.TASKS_INPUTS_RSA_PRIVATE_KEY = b"input_private_key"
        mock_settings.TASKS_INPUTS_ECDSA_PUBLIC_KEY = b"input_public_key"
        mock_settings.TASKS_OUTPUTS_RSA_PUBLIC_KEY = b"output_public_key"
        mock_settings.TASKS_OUTPUTS_ECDSA_PRIVATE_KEY = b"output_private_key"

        task = Task(
            name="test_task",
            content="encrypted_content",
            content_metadata="input_metadata",
            result=None
        )
        original_content = task.content
        decrypted_content = "decrypted_content"
        encrypted_result = "encrypted_result"
        result_metadata = "result_metadata"

        mock_decrypt = mocker.patch(
            "octobot_node.scheduler.task_context.decrypt_task_content"
        )
        mock_decrypt.return_value = decrypted_content

        mock_encrypt = mocker.patch(
            "octobot_node.scheduler.task_context.encrypt_task_result"
        )
        mock_encrypt.return_value = (encrypted_result, result_metadata)

        mock_json_dumps = mocker.patch(
            "octobot_node.scheduler.task_context.json.dumps"
        )
        mock_json_dumps.return_value = '{"status": "success"}'

        with encrypted_task(task):
            # Content should be decrypted
            assert task.content == decrypted_content
            # Set result inside context
            task.result = {"status": "success"}

        # Content should be restored
        assert task.content == original_content
        # Result should be encrypted
        assert task.result == encrypted_result
        assert task.result_metadata == result_metadata

    def test_encrypted_task_decryption_error(
        self, mocker: MockerFixture
    ) -> None:
        mock_settings = mocker.patch("octobot_node.scheduler.task_context.settings")
        mock_settings.TASKS_INPUTS_RSA_PRIVATE_KEY = b"private_key"
        mock_settings.TASKS_INPUTS_ECDSA_PUBLIC_KEY = b"public_key"
        mock_settings.TASKS_OUTPUTS_RSA_PUBLIC_KEY = None
        mock_settings.TASKS_OUTPUTS_ECDSA_PRIVATE_KEY = None

        task = Task(
            name="test_task",
            content="encrypted_content",
            content_metadata="metadata",
            result=None
        )
        original_content = task.content
        decryption_error = ValueError("Decryption failed")

        mock_decrypt = mocker.patch(
            "octobot_node.scheduler.task_context.decrypt_task_content"
        )
        mock_decrypt.side_effect = decryption_error

        mock_logger = mocker.patch("octobot_node.scheduler.task_context.logger")

        with encrypted_task(task):
            # Content should remain unchanged on error
            assert task.content == original_content
            # Set a result
            task.result = {"status": "success"}

        # Content should still be original
        assert task.content == original_content
        # Result should be set to error format
        assert task.result[TaskResultKeys.STATUS.value] == "failed"
        assert task.result[TaskResultKeys.TASK.value] == {"name": "test_task"}
        assert task.result[TaskResultKeys.RESULT.value] == {}
        assert task.result[TaskResultKeys.ERROR.value] == str(decryption_error)
        mock_logger.error.assert_called_once()

    def test_encrypted_task_result_none(
        self, mocker: MockerFixture
    ) -> None:
        mock_settings = mocker.patch("octobot_node.scheduler.task_context.settings")
        mock_settings.TASKS_INPUTS_RSA_PRIVATE_KEY = None
        mock_settings.TASKS_INPUTS_ECDSA_PUBLIC_KEY = None
        mock_settings.TASKS_OUTPUTS_RSA_PUBLIC_KEY = b"public_key"
        mock_settings.TASKS_OUTPUTS_ECDSA_PRIVATE_KEY = b"private_key"

        task = Task(
            name="test_task",
            content="plain content",
            result=None
        )

        mock_encrypt = mocker.patch(
            "octobot_node.scheduler.task_context.encrypt_task_result"
        )

        with encrypted_task(task):
            # Don't set result
            pass

        # Encryption should not be called when result is None
        mock_encrypt.assert_not_called()

    def test_encrypted_task_exception_during_context(
        self, mocker: MockerFixture
    ) -> None:
        mock_settings = mocker.patch("octobot_node.scheduler.task_context.settings")
        mock_settings.TASKS_INPUTS_RSA_PRIVATE_KEY = b"private_key"
        mock_settings.TASKS_INPUTS_ECDSA_PUBLIC_KEY = b"public_key"
        mock_settings.TASKS_OUTPUTS_RSA_PUBLIC_KEY = None
        mock_settings.TASKS_OUTPUTS_ECDSA_PRIVATE_KEY = None

        task = Task(
            name="test_task",
            content="encrypted_content",
            content_metadata="metadata"
        )
        original_content = task.content
        decrypted_content = "decrypted_content"

        mock_decrypt = mocker.patch(
            "octobot_node.scheduler.task_context.decrypt_task_content"
        )
        mock_decrypt.return_value = decrypted_content

        # Exception should propagate, but content should be restored
        with pytest.raises(ValueError, match="Test exception"):
            with encrypted_task(task):
                # Content should be decrypted
                assert task.content == decrypted_content
                # Raise exception
                raise ValueError("Test exception")

        # Content should be restored even after exception
        assert task.content == original_content



#  This file is part of OctoBot Node (https://github.com/Drakkar-Software/OctoBot-Node)
#  Copyright (c) 2025 Drakkar-Software, All rights reserved.
#
#  OctoBot is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  OctoBot is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#
#  You should have received a copy of the GNU General Public
#  License along with OctoBot. If not, see <https://www.gnu.org/licenses/>.

import contextlib
import logging
import json
from typing import Any

from octobot_node.scheduler.encryption import decrypt_task_content, encrypt_task_result
from octobot_node.app.core.config import settings
from octobot_node.app.models import Task
from octobot_node.app.enums import TaskResultKeys


logger = logging.getLogger(__name__)


@contextlib.contextmanager
def encrypted_task(task: Task):
    """
    Context manager for automatically decrypting task content and encrypting task results.
    It first decrypts task.content if TASKS_INPUTS_RSA_PRIVATE_KEY is provided.
    Then it encrypts the result and restores original content on exit if TASKS_OUTPUTS_RSA_PUBLIC_KEY is provided
    """
    original_content = task.content
    decryption_error: Exception | None = None
    
    try:
        # Decrypt content if encryption keys are configured
        if settings.TASKS_INPUTS_RSA_PRIVATE_KEY and settings.TASKS_INPUTS_ECDSA_PUBLIC_KEY:
            try:
                decrypted_content = decrypt_task_content(task.content, task.content_metadata)
                task.content = decrypted_content
            except Exception as e:
                logger.error(f"Failed to decrypt content: {e}")
                decryption_error = e
        
        yield task
    finally:
        # Restore original content if it was modified
        if task.content != original_content:
            task.content = original_content

        if task.result is not None:
            if decryption_error:
                task.result = {
                    TaskResultKeys.STATUS.value: "failed", 
                    TaskResultKeys.TASK.value: {"name": task.name}, 
                    TaskResultKeys.RESULT.value: {}, 
                    TaskResultKeys.ERROR.value: str(decryption_error)
                }

            # Encrypt result if encryption keys are configured
            if settings.TASKS_OUTPUTS_RSA_PUBLIC_KEY and settings.TASKS_OUTPUTS_ECDSA_PRIVATE_KEY:
                try:
                    result_json = json.dumps(task.result)
                    encrypted_result, metadata = encrypt_task_result(result_json)
                    task.result = encrypted_result
                    task.result_metadata = metadata
                except Exception as e:
                    logger.error(f"Failed to encrypt result: {e}")
                    # TODO: Handle encryption failure



#  This file is part of OctoBot Node (https://github.com/Drakkar-Software/OctoBot-Node)
#  Copyright (c) 2025 Drakkar-Software, All rights reserved.
#
#  OctoBot is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  OctoBot is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#
#  You should have received a copy of the GNU General Public
#  License along with OctoBot. If not, see <https://www.gnu.org/licenses/>.

from octobot_node.scheduler import SCHEDULER
from octobot_node.scheduler.task_context import encrypted_task
from octobot_node.app.models import Task, TaskType
from octobot_node.app.enums import TaskResultKeys
from octobot_node.app.models import TaskStatus


@SCHEDULER.INSTANCE.task()
def start_octobot(task: Task):
    with encrypted_task(task):
        # TODO
        task.result = "ok"
    return {
        TaskResultKeys.STATUS.value: TaskStatus.COMPLETED.value,
        TaskResultKeys.RESULT.value: task.result, 
        TaskResultKeys.METADATA.value: task.result_metadata,
        TaskResultKeys.TASK.value: {"name": task.name},
        TaskResultKeys.ERROR.value: None
    }


@SCHEDULER.INSTANCE.task()
def execute_octobot(task: Task):
    with encrypted_task(task):
        if task.type == TaskType.EXECUTE_ACTIONS.value:
            # TODO start_octobot with actions
            print(f"Executing actions with content: {task.content}...")
            task.result = {
                "state": {
                    "orders": [], # WIP
                }
            }
        else:
            raise ValueError(f"Invalid task type: {task.type}")
    return {
        TaskResultKeys.STATUS.value: TaskStatus.COMPLETED.value,
        TaskResultKeys.RESULT.value: task.result, 
        TaskResultKeys.METADATA.value: task.result_metadata,
        TaskResultKeys.TASK.value: {"name": task.name},
        TaskResultKeys.ERROR.value: None
    }


@SCHEDULER.INSTANCE.task()
def stop_octobot(task: Task):
    with encrypted_task(task):
        # TODO
        task.result = "ok"
    return {
        TaskResultKeys.STATUS.value: TaskStatus.COMPLETED.value,
        TaskResultKeys.RESULT.value: task.result, 
        TaskResultKeys.METADATA.value: task.result_metadata,
        TaskResultKeys.TASK.value: {"name": task.name},
        TaskResultKeys.ERROR.value: None
    }

def trigger_task(task: Task) -> bool:
    if task.type == TaskType.START_OCTOBOT.value:
        start_octobot.schedule(args=[task], delay=1)
        return True
    elif task.type == TaskType.STOP_OCTOBOT.value:
        stop_octobot.schedule(args=[task], delay=1)
        return True
    elif task.type == TaskType.EXECUTE_ACTIONS.value:
        execute_octobot.schedule(args=[task], delay=1)
        return True
    else:
        raise ValueError(f"Invalid task type: {task.type}")



#  This file is part of OctoBot Node (https://github.com/Drakkar-Software/OctoBot-Node)
#  Copyright (c) 2025 Drakkar-Software, All rights reserved.
#
#  OctoBot is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  OctoBot is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#
#  You should have received a copy of the GNU General Public
#  License along with OctoBot. If not, see <https://www.gnu.org/licenses/>.

# Encryption tasks constants
ENCRYPTED_AES_KEY_B64_METADATA_KEY = "ENCRYPTED_AES_KEY_B64"
IV_B64_METADATA_KEY = "IV_B64"
SIGNATURE_B64_METADATA_KEY = "SIGNATURE_B64"

# Encryption tasks errors
class MissingMetadataError(Exception):
    pass

class MetadataParsingError(Exception):
    pass

class EncryptionTaskError(Exception):
    pass

class SignatureVerificationError(Exception):
    pass

from octobot_node.scheduler.encryption import task_inputs
from octobot_node.scheduler.encryption.task_inputs import (decrypt_task_content, encrypt_task_content)

from octobot_node.scheduler.encryption import task_outputs
from octobot_node.scheduler.encryption.task_outputs import (encrypt_task_result, decrypt_task_result)

__all__ = [
    "ENCRYPTED_AES_KEY_B64_METADATA_KEY",
    "IV_B64_METADATA_KEY",
    "SIGNATURE_B64_METADATA_KEY",
    "MissingMetadataError",
    "MetadataParsingError",
    "EncryptionTaskError",
    "SignatureVerificationError",
    "decrypt_task_content",
    "encrypt_task_content",
    "encrypt_task_result",
    "decrypt_task_result"
]


#  This file is part of OctoBot Node (https://github.com/Drakkar-Software/OctoBot-Node)
#  Copyright (c) 2025 Drakkar-Software, All rights reserved.
#
#  OctoBot is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  OctoBot is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#
#  You should have received a copy of the GNU General Public
#  License along with OctoBot. If not, see <https://www.gnu.org/licenses/>.

import logging
import threading

from octobot_node.app.core.config import settings

from octobot_node.scheduler.scheduler import Scheduler
from huey.constants import WORKER_THREAD

class SchedulerConsumer:
    def __init__(self, scheduler: Scheduler):
        self.scheduler: Scheduler = scheduler
        self.consumer = None
        self.thread = None
        self.lock = threading.Lock()
        self.logger = logging.getLogger(__name__)
        self.workers: int | None = None
    
    def start(self):
        with self.lock:
            if self.thread is None or not self.thread.is_alive():
                self.start_thread()
                self.logger.info("Scheduler consumer thread started automatically on import")
    
    def _run(self):
        if not self.consumer:
            self.logger.error("Consumer not initialized. Cannot start consumer thread.")
            return
        self.logger.info("Starting consumer...")
        try:
            self.consumer.run()
        except ValueError as e:
            # Ignore `ValueError: signal only works in main thread of the main interpreter``
            self.logger.debug(f"ValueError ignored when starting consumer: {e}")

    def start_thread(self) -> None:
        if settings.SCHEDULER_WORKERS <= 0:
            self.logger.info("Consumers are disabled.")
            self.workers = None
            return
        self.logger.info(f"Starting {settings.SCHEDULER_WORKERS} scheduler consumer")
        self.workers = settings.SCHEDULER_WORKERS
        config_values = {
            "worker_type": WORKER_THREAD,
            "workers": self.workers,
        }
        self.consumer = self.scheduler.INSTANCE.create_consumer(**config_values)
        self.logger.info(
            f"Scheduler consumer started with {self.workers} workers (worker_type=thread)",
            self.workers
        )
        self.thread = threading.Thread(
            target=self._run,
            args=(),
            name="scheduler-consumer",
            daemon=True,
        )
        self.thread.start()
        self.logger.info("Scheduler consumer running in thread")

    def is_started(self) -> bool:
        with self.lock:
            return self.thread is not None

    def is_running(self) -> bool:
        with self.lock:
            return self.thread is not None and self.thread.is_alive()

    def stop(self) -> None:
        with self.lock:
            if not self.consumer:
                return

            self.consumer.stop(graceful=True)
            if self.thread:
                self.thread.join(timeout=5)

            self.logger.info("Scheduler consumer stopped")
            self.consumer = None
            self.thread = None
            self.workers = None



#  This file is part of OctoBot Node (https://github.com/Drakkar-Software/OctoBot-Node)
#  Copyright (c) 2025 Drakkar-Software, All rights reserved.
#
#  OctoBot is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  OctoBot is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#
#  You should have received a copy of the GNU General Public
#  License along with OctoBot. If not, see <https://www.gnu.org/licenses/>.

import json
import base64

from typing import Optional, Tuple
from octobot_node.app.core.config import settings
from octobot_node.scheduler.encryption import (
    ENCRYPTED_AES_KEY_B64_METADATA_KEY, 
    IV_B64_METADATA_KEY, 
    SIGNATURE_B64_METADATA_KEY, 
    MissingMetadataError, 
    EncryptionTaskError, 
    MetadataParsingError, 
    SignatureVerificationError
)
import octobot_commons.cryptography as cryptography

def decrypt_task_content(content: str, metadata: Optional[str] = None) -> str:
    if metadata is None:
        raise MissingMetadataError("No metadata provided for content decryption")

    try:
        metadata = json.loads(base64.b64decode(metadata).decode('utf-8'))
        encrypted_aes_key_b64 = metadata.get(ENCRYPTED_AES_KEY_B64_METADATA_KEY, None)
        iv_b64 = metadata.get(IV_B64_METADATA_KEY, None)
        signature_b64 = metadata.get(SIGNATURE_B64_METADATA_KEY, None)
    except Exception as e:
        raise MetadataParsingError(f"Failed to parse encrypted AES key or IV from metadata: {e}")

    if not encrypted_aes_key_b64 or not iv_b64 or not signature_b64:
        raise MissingMetadataError("No encrypted AES key or IV or signature provided for content decryption")

    try:
        content_bytes = base64.b64decode(content)
        encrypted_aes_key = base64.b64decode(encrypted_aes_key_b64)
        iv = base64.b64decode(iv_b64)
        signature = base64.b64decode(signature_b64)
    except Exception as e:
        raise MetadataParsingError(f"Failed to decode base64-encoded data: {e}")

    data_to_verify = content_bytes + encrypted_aes_key + iv
    if not cryptography.verify_signature(data_to_verify, settings.TASKS_INPUTS_ECDSA_PUBLIC_KEY, signature):
        raise SignatureVerificationError("Signature verification failed")

    decrypted_aes_key = cryptography.rsa_decrypt_aes_key(encrypted_aes_key, settings.TASKS_INPUTS_RSA_PRIVATE_KEY)
    if not decrypted_aes_key:
        raise EncryptionTaskError("Failed to decrypt AES key")

    decrypted_content = cryptography.aes_gcm_decrypt(content_bytes, decrypted_aes_key, iv)
    if not decrypted_content:
        raise EncryptionTaskError("Failed to decrypt content")

    return decrypted_content.decode('utf-8')


def encrypt_task_content(content: str) -> Tuple[str, str]:
    aes_encryption_key = cryptography.generate_aes_key()
    iv = cryptography.generate_iv()

    encrypted_content = cryptography.aes_gcm_encrypt(content.encode('utf-8'), aes_encryption_key, iv)
    if not encrypted_content:
        raise EncryptionTaskError("Failed to encrypt content")

    encrypted_aes_key = cryptography.rsa_encrypt_aes_key(aes_encryption_key, settings.TASKS_INPUTS_RSA_PUBLIC_KEY)
    if not encrypted_aes_key:
        raise EncryptionTaskError("Failed to encrypt AES key")

    data_to_sign = encrypted_content + encrypted_aes_key + iv
    signature = cryptography.sign_data(data_to_sign, settings.TASKS_INPUTS_ECDSA_PRIVATE_KEY)
    if not signature:
        raise EncryptionTaskError("Failed to sign data")

    metadata = {
        ENCRYPTED_AES_KEY_B64_METADATA_KEY: base64.b64encode(encrypted_aes_key).decode('utf-8'),
        IV_B64_METADATA_KEY: base64.b64encode(iv).decode('utf-8'),
        SIGNATURE_B64_METADATA_KEY: base64.b64encode(signature).decode('utf-8'),
    }
    encrypted_content_b64 = base64.b64encode(encrypted_content).decode('utf-8')
    metadata_json = json.dumps(metadata)
    metadata_b64 = base64.b64encode(metadata_json.encode('utf-8')).decode('utf-8')
    return encrypted_content_b64, metadata_b64



#  This file is part of OctoBot Node (https://github.com/Drakkar-Software/OctoBot-Node)
#  Copyright (c) 2025 Drakkar-Software, All rights reserved.
#
#  OctoBot is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  OctoBot is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#
#  You should have received a copy of the GNU General Public
#  License along with OctoBot. If not, see <https://www.gnu.org/licenses/>.
import argparse
import sys

try:
    import uvicorn
except ImportError:
    print(
        "Error importing uvicorn, please install OctoBot-Node with all dependencies. "
        "Example: \"pip install -U octobot-node\" "
        "(Error: uvicorn not found)", file=sys.stderr
    )
    sys.exit(-1)

from octobot_node import PROJECT_NAME, LONG_VERSION


def start_server(args):
    port = args.port or 8000
    
    # This must be done before the scheduler module is imported
    from octobot_node.app.core.config import settings
    
    if args.master:
        settings.IS_MASTER_MODE = True
    settings.SCHEDULER_WORKERS = args.consumers

    # Check that the node is either a master node or has consumers enabled
    if not settings.IS_MASTER_MODE and settings.SCHEDULER_WORKERS <= 0:
        print(
            "Error: Node must be either a master node (--master) or have consumers enabled (--consumers N).\n"
            "  - Use --master to enable master node mode (schedules tasks)\n"
            "  - Use --consumers N to enable consumer workers (processes tasks)\n"
            "  - Use both --master --consumers N to enable both modes",
            file=sys.stderr
        )
        sys.exit(1)

    from octobot_node.scheduler import CONSUMER
    
    if args.environment is not None:
        settings.ENVIRONMENT = args.environment
    
    if args.admin_username is not None:
        settings.ADMIN_USERNAME = args.admin_username
    
    if args.admin_password is not None:
        settings.ADMIN_PASSWORD = args.admin_password
    
    # If not master mode, bind to localhost only
    if args.host is not None:
        host = args.host
    elif settings.IS_MASTER_MODE and settings.ENVIRONMENT == "production":
        host = "0.0.0.0"
    else:
        host = "127.0.0.1"

    # Ensure settings are loaded
    from octobot_node.app.main import app
    
    uvicorn.run(
        app,
        host=host,
        port=port,
        log_level="info",
        access_log=args.verbose,
    )


def octobot_node_parser(parser):
    """Configure argument parser for OctoBot-Node CLI."""
    parser.add_argument(
        '-v', '--version',
        help='Show OctoBot-Node current version.',
        action='store_true'
    )
    parser.add_argument(
        '--host',
        help='Host to bind the server to (default: 0.0.0.0).',
        type=str,
        default=None
    )
    parser.add_argument(
        '--port',
        help='Port to bind the server to (default: 8000).',
        type=int,
        default=None
    )
    parser.add_argument(
        '--master',
        help='Enable master node mode (schedules tasks, UI enabled by default).',
        action='store_true'
    )
    parser.add_argument(
        '--consumers',
        help='Number of consumer worker threads (0 disables consumers, default: 0). Can be used with --master.',
        type=int,
        default=0
    )
    parser.add_argument(
        '--environment',
        help='Environment mode: local or production (default: from ENVIRONMENT environment variable). Auto-reload is enabled when environment is local.',
        type=str,
        choices=['local', 'production'],
        default=None
    )
    parser.add_argument(
        '--admin-username',
        help='Admin username (email format). Default: from ADMIN_USERNAME environment variable.',
        type=str,
        default=None
    )
    parser.add_argument(
        '--admin-password',
        help='Admin password. Default: from ADMIN_PASSWORD environment variable.',
        type=str,
        default=None
    )
    parser.add_argument(
        '--verbose',
        help='Enable verbose logging, including HTTP access logs.',
        action='store_true'
    )
    parser.set_defaults(func=start_octobot_node)


def start_octobot_node(args):
    if args.version:
        print(LONG_VERSION)
        return
    
    start_server(args)


def main(args=None):
    if not args:
        args = sys.argv[1:]
    
    parser = argparse.ArgumentParser(
        description=PROJECT_NAME,
        formatter_class=argparse.RawDescriptionHelpFormatter
    )
    octobot_node_parser(parser)
    
    parsed_args = parser.parse_args(args)
    parsed_args.func(parsed_args)




#  This file is part of OctoBot Node (https://github.com/Drakkar-Software/OctoBot-Node)
#  Copyright (c) 2025 Drakkar-Software, All rights reserved.
#
#  OctoBot is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  OctoBot is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#
#  You should have received a copy of the GNU General Public
#  License along with OctoBot. If not, see <https://www.gnu.org/licenses/>.

from huey import Huey, RedisHuey, SqliteHuey
from huey.registry import Message
from huey.utils import Error as HueyError
from typing import Optional, Any
import logging
import pickle
import json
from octobot_node.app.models import Task, TaskType, TaskStatus
from octobot_node.app.core.config import settings
from octobot_node.app.enums import TaskResultKeys

DEFAULT_NAME = "octobot_node"

class Scheduler:
    INSTANCE: Optional[Huey] = None

    def __init__(self):
        self.logger = logging.getLogger(self.__class__.__name__)

    def create(self):
        if settings.SCHEDULER_REDIS_URL:
            import redis
            self.logger.info(
                "Initializing scheduler with Redis backend at %s", settings.SCHEDULER_REDIS_URL
            )

            connection_pool = redis.ConnectionPool.from_url(
                str(settings.SCHEDULER_REDIS_URL),
                ssl_ca_certs=f"{settings.REDIS_STORAGE_CERTS_PATH}/ca.crt",
                ssl_certfile=f"{settings.REDIS_STORAGE_CERTS_PATH}/client/client.crt",
                ssl_keyfile=f"{settings.REDIS_STORAGE_CERTS_PATH}/client/client.key",
                ssl_cert_reqs="required",
                decode_responses=False,
                socket_timeout=5,
                socket_connect_timeout=3
            ) if settings.REDIS_STORAGE_CERTS_PATH is not None else None

            self.INSTANCE = RedisHuey(DEFAULT_NAME, connection_pool=connection_pool) if connection_pool is not None else RedisHuey(DEFAULT_NAME, url=str(settings.SCHEDULER_REDIS_URL))
        else:
            self.logger.info(
                "Initializing scheduler with sqlite backend at %s", settings.SCHEDULER_SQLITE_FILE
            )
            self.INSTANCE = SqliteHuey(DEFAULT_NAME, filename=settings.SCHEDULER_SQLITE_FILE)

    def stop(self) -> None:
        if self.INSTANCE:
            # TODO self.INSTANCE.stop()
            self.logger.info("Scheduler stopped")
        else:
            self.logger.warning("Scheduler not initialized")

    def get_periodic_tasks(self) -> list[dict]:
        tasks: list[dict] = []
        periodic_tasks = self.INSTANCE._registry.periodic_tasks
        for task in periodic_tasks or []:
            try:
                tasks.append(self._parse_task(task, TaskStatus.PERIODIC, f"Periodic task: {task.name}"))
            except Exception as e:
                self.logger.warning(f"Failed to process periodic task {task.name}: {e}")
        return tasks

    def get_pending_tasks(self) -> list[dict]:
        tasks: list[dict] = []
        pending_tasks = self.INSTANCE.pending()
        for task in pending_tasks or []:
            try:
                tasks.append(self._parse_task(task, TaskStatus.PENDING, f"Pending task: {task.name}"))
            except Exception as e:
                self.logger.warning(f"Failed to process pending task {task.name}: {e}")
        return tasks

    def get_scheduled_tasks(self) -> list[dict]:
        tasks: list[dict] = []
        scheduled_tasks = self.INSTANCE.scheduled()
        for task in scheduled_tasks or []:
            try:
                tasks.append(self._parse_task(task, TaskStatus.SCHEDULED, f"Scheduled at {task.eta.strftime('%Y-%m-%d %H:%M:%S')}"))
            except Exception as e:
                self.logger.warning(f"Failed to process scheduled task {task.name}: {e}")
        return tasks

    def _decode_result(self, result_key_bytes: bytes | str, result_value_bytes: bytes | Any) -> tuple[str, Any | None]:
        task_id = result_key_bytes.decode('utf-8') if isinstance(result_key_bytes, bytes) else result_key_bytes
        
        try:
            result_obj = pickle.loads(result_value_bytes) if isinstance(result_value_bytes, bytes) else result_value_bytes
            return (task_id, result_obj)
        except (pickle.UnpicklingError, Exception) as unpickle_error:
            self.logger.warning(f"Failed to unpickle result for task {task_id}: {unpickle_error}")
            return (task_id, None)

    def get_results(self) -> list[dict]:
        tasks: list[dict] = []
        result_keys = self.INSTANCE.all_results()
        for result_key_bytes, result_value_bytes in result_keys.items():
            try:
                task_id, result_obj = self._decode_result(result_key_bytes, result_value_bytes)
                
                if result_obj is None:
                    description = f"Task completed (unable to parse result)"
                    status = TaskStatus.COMPLETED
                elif isinstance(result_obj, HueyError):
                    description = f"Task failed: {result_obj.error}"
                    status = TaskStatus.FAILED
                else:
                    description = f"Task completed"
                    status = TaskStatus.COMPLETED
                
                result = result_obj.get(TaskResultKeys.RESULT.value)
                metadata = result_obj.get(TaskResultKeys.METADATA.value)

                tasks.append({
                    "id": task_id,
                    "name": self.get_task_name(result_obj, task_id),
                    "description": description,
                    "status": status,
                    "result": json.dumps(result),
                    "result_metadata": metadata,
                    "scheduled_at": None,
                    "started_at": None,
                    "completed_at": None,
                })
            except Exception as e:
                self.logger.warning(f"Failed to process result key {result_key_bytes}: {e}")
        return tasks


    def _parse_task(self, message: Message, status: TaskStatus, description: Optional[str] = None) -> Task:
        task_kwargs = message.kwargs
        task_args = message.args
        task_actions = task_kwargs.get("actions")
        task_type = task_kwargs.get("type")
        task_name = self.get_task_name(task_args[0] if task_args and len(task_args) > 0 else {}, message.name)

        return Task(
            id=message.id,
            name=task_name,
            description=description,
            actions=task_actions,
            type=TaskType(task_type) if task_type else None,
            status=status,
            retries=message.retries,
            retry_delay=message.retry_delay,
            priority=message.priority,
            expires=message.expires,
            expires_resolved=message.expires_resolved,
            scheduled_at=message.eta,
            started_at=None,
            completed_at=None,
        )


    def save_data(self, key: str, value: str) -> None:
        self.INSTANCE.storage.put_data(key, value)

    def get_data(self, key: str) -> str:
        return self.INSTANCE.storage.peek_data(key)

    def get_task_name(self, task_data: dict | Task | None, default_value: Optional[str] = None) -> Optional[str]:
        if isinstance(task_data, Task):
            return task_data.name
        elif isinstance(task_data, dict):
            return task_data.get(TaskResultKeys.TASK.value, {}).get("name", default_value)
        else:
            return default_value



#  This file is part of OctoBot Node (https://github.com/Drakkar-Software/OctoBot-Node)
#  Copyright (c) 2025 Drakkar-Software, All rights reserved.
#
#  OctoBot is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  OctoBot is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#
#  You should have received a copy of the GNU General Public
#  License along with OctoBot. If not, see <https://www.gnu.org/licenses/>.

import logging
import uuid
from typing import Any

from octobot_node.app.core.config import settings
from octobot_node.scheduler import SCHEDULER, CONSUMER

logger = logging.getLogger(__name__)


def get_node_status() -> dict[str, str | int | None | uuid.UUID]:
    consumer_running = CONSUMER.is_started() # TODO use: CONSUMER.is_running()
    is_running = settings.IS_MASTER_MODE or consumer_running
    status = "running" if is_running else "stopped"

    backend_type = "redis" if settings.SCHEDULER_REDIS_URL else "sqlite"
    workers = CONSUMER.workers if settings.SCHEDULER_WORKERS > 0 else None

    if settings.IS_MASTER_MODE and settings.SCHEDULER_WORKERS > 0:
        node_type = "both"
    elif settings.IS_MASTER_MODE:
        node_type = "master"
    elif settings.SCHEDULER_WORKERS > 0:
        node_type = "consumer"
    else:
        node_type = "none"

    return {
        "node_type": node_type,
        "backend_type": backend_type,
        "workers": workers,
        "status": status,
        "redis_url": str(settings.SCHEDULER_REDIS_URL) if settings.SCHEDULER_REDIS_URL else None,
        "sqlite_file": settings.SCHEDULER_SQLITE_FILE if not settings.SCHEDULER_REDIS_URL else None,
    }


def get_task_metrics() -> dict[str, int]:
    try:
        huey_instance = SCHEDULER.INSTANCE
        if huey_instance is None:
            logger.warning("Scheduler instance not initialized")
            return {"pending": 0, "scheduled": 0, "results": 0}

        scheduled_count = huey_instance.scheduled_count()
        periodic_tasks = SCHEDULER.get_periodic_tasks()
        scheduled_count += len(periodic_tasks)

        return {
            "pending": huey_instance.pending_count(),
            "scheduled": scheduled_count,
            "results": huey_instance.result_count(),
        }
    except Exception as e:
        logger.error("Failed to retrieve task metrics from scheduler: %s", e)
        return {"pending": 0, "scheduled": 0, "results": 0}


def get_all_tasks() -> list[dict[str, Any]]:
    tasks: list[dict[str, Any]] = []
    try:
        periodic_tasks = SCHEDULER.get_periodic_tasks()
        tasks.extend(periodic_tasks)

        pending_tasks = SCHEDULER.get_pending_tasks()
        tasks.extend(pending_tasks)

        scheduled_tasks = SCHEDULER.get_scheduled_tasks()
        tasks.extend(scheduled_tasks)

        results = SCHEDULER.get_results()
        tasks.extend(results)
    except Exception as e:
        logger.error("Failed to retrieve tasks from scheduler: %s", e)

    logger.debug("Returning %d total tasks", len(tasks))
    return tasks


async def get_task_result(task_id: str):
    res = SCHEDULER.INSTANCE.result(task_id)

    if res is None:
        return {"error": "task not found"}

    # True if finished
    if res.ready():          
        # blocks if not ready, or returns the value                        
        result_data = res.get()
        return {"status": "completed", "data": result_data}
    else:
        return {"status": "pending or running"}



#  This file is part of OctoBot Node (https://github.com/Drakkar-Software/OctoBot-Node)
#  Copyright (c) 2025 Drakkar-Software, All rights reserved.
#
#  OctoBot is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  OctoBot is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#
#  You should have received a copy of the GNU General Public
#  License along with OctoBot. If not, see <https://www.gnu.org/licenses/>.

from contextlib import asynccontextmanager

import sentry_sdk
from fastapi import FastAPI, HTTPException, Request
from fastapi.responses import FileResponse
from fastapi.routing import APIRoute
from fastapi.staticfiles import StaticFiles
from starlette.middleware.cors import CORSMiddleware

from octobot_node import PROJECT_NAME
from octobot_node.app.api.main import api_router
from octobot_node.app.core.config import settings
from octobot_node.app.utils import get_dist_directory
from octobot_node.scheduler import SCHEDULER, CONSUMER


def custom_generate_unique_id(route: APIRoute) -> str:
    if route.tags:
        return f"{route.tags[0]}-{route.name}"
    # Fallback for routes without tags (e.g., SPA root)
    return route.name or route.path.replace("/", "-").strip("-")


if settings.SENTRY_DSN and settings.ENVIRONMENT != "local":
    sentry_sdk.init(
        dsn=str(settings.SENTRY_DSN),
        enable_tracing=True,
        include_local_variables=False,   # careful not to upload sensitive data
    )


@asynccontextmanager
async def lifespan(app: FastAPI):
    """Manage application lifespan: startup and shutdown events."""
    # Startup - scheduler starts automatically on import
    # Import scheduler module to ensure it's initialized
    from octobot_node.scheduler import scheduler  # noqa: F401
    yield
    # Shutdown
    SCHEDULER.stop()
    CONSUMER.stop()


app = FastAPI(
    title=PROJECT_NAME,
    openapi_url=f"{settings.API_V1_STR}/openapi.json",
    generate_unique_id_function=custom_generate_unique_id,
    lifespan=lifespan,
)

# Set all CORS enabled origins
if settings.all_cors_origins:
    app.add_middleware(
        CORSMiddleware,
        allow_origins=settings.all_cors_origins,
        allow_credentials=True,
        allow_methods=["*"],
        allow_headers=["*"],
    )

app.include_router(api_router, prefix=settings.API_V1_STR)

# Get the path to the dist folder (works for both development and installed packages)
DIST_DIR = get_dist_directory()

# Serve static files from the dist folder only if UI is enabled
if DIST_DIR:
    assets_dir = DIST_DIR / "assets"
    if assets_dir.exists():
        # Mount assets under /app/assets to match the SPA base path
        app.mount("/app/assets", StaticFiles(directory=str(assets_dir)), name="assets")
    
    # Serve SPA root for /app
    @app.get("/app")
    async def serve_spa_app_root():
        index_path = DIST_DIR / "index.html"
        if index_path.exists():
            return FileResponse(str(index_path))
        raise HTTPException(status_code=404, detail="Frontend build not found")
    
    # Serve SPA for /app routes
    @app.get("/app/{path:path}")
    async def serve_spa_app(request: Request, path: str):
        """
        Serve the React app for /app routes.
        This enables client-side routing.
        """
        # Don't interfere with assets (already handled by mount)
        if path.startswith("assets/"):
            raise HTTPException(status_code=404)
        
        # Serve index.html for all /app routes (SPA routing)
        index_path = DIST_DIR / "index.html"
        if index_path.exists():
            return FileResponse(str(index_path))
        raise HTTPException(status_code=404, detail="Frontend build not found")



#  Drakkar-Software OctoBot-Node
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

import csv
import json
import os
from typing import Dict, List, Optional

import octobot_commons.cryptography as cryptography
from octobot_node.scheduler.encryption.task_inputs import (
    encrypt_task_content,
    decrypt_task_content,
)

################################################################################
# This file is used to test the functions inside octobot_node/ui/src/lib/csv.ts
# It provides Python implementations of the CSV column merging logic to verify
# that the TypeScript implementation behaves correctly.
################################################################################

COLUMN_NAME = "name"
COLUMN_CONTENT = "content"
COLUMN_TYPE = "type"
CONTENT_SEPARATOR = ";"

REQUIRED_KEYS = [COLUMN_NAME, COLUMN_TYPE]
KEYS_OUTSIDE_CONTENT = [COLUMN_NAME, COLUMN_TYPE]

DEFAULT_KEYS_FILE = "task_encryption_keys.json"

KEY_NAMES = {
    "TASKS_INPUTS_RSA_PUBLIC_KEY": "tasks_inputs_rsa_public_key",
    "TASKS_INPUTS_RSA_PRIVATE_KEY": "tasks_inputs_rsa_private_key",
    "TASKS_INPUTS_ECDSA_PUBLIC_KEY": "tasks_inputs_ecdsa_public_key",
    "TASKS_INPUTS_ECDSA_PRIVATE_KEY": "tasks_inputs_ecdsa_private_key",
    "TASKS_OUTPUTS_RSA_PUBLIC_KEY": "tasks_outputs_rsa_public_key",
    "TASKS_OUTPUTS_RSA_PRIVATE_KEY": "tasks_outputs_rsa_private_key",
    "TASKS_OUTPUTS_ECDSA_PUBLIC_KEY": "tasks_outputs_ecdsa_public_key",
    "TASKS_OUTPUTS_ECDSA_PRIVATE_KEY": "tasks_outputs_ecdsa_private_key",
}


def find_column_index(column_names: List[str], key: str) -> int:
    for i, col in enumerate(column_names):
        if col.lower() == key.lower():
            return i
    return -1


def validate_required_keys(column_names: List[str]) -> Dict[int, str]:
    required_keys_indices: Dict[int, str] = {}
    for key in REQUIRED_KEYS:
        index = find_column_index(column_names, key)
        if index == -1:
            raise ValueError(f"Required key '{key}' not found in CSV header")
        required_keys_indices[index] = key
    return required_keys_indices


def find_keys_outside_content_indices(column_names: List[str]) -> Dict[int, str]:
    indices: Dict[int, str] = {}
    for key in KEYS_OUTSIDE_CONTENT:
        index = find_column_index(column_names, key)
        if index != -1:
            indices[index] = key
    return indices


def build_content(
    values: List[str],
    column_names: List[str],
    keys_outside_content_indices: Dict[int, str],
    content_column_index: int
) -> str:
    content_parts: List[str] = []
    
    for i in range(min(len(column_names), len(values))):
        if i not in keys_outside_content_indices and i != content_column_index:
            value = values[i].strip() if i < len(values) else ""
            if value:
                column_name = column_names[i]
                upper_key = column_name.upper()
                content_parts.append(f"{upper_key}={value}")
    
    concatenated_content = CONTENT_SEPARATOR.join(content_parts)
    if content_column_index != -1 and content_column_index < len(values):
        content_column_value = values[content_column_index].strip()
        if content_column_value:
            if concatenated_content:
                return f"{concatenated_content}{CONTENT_SEPARATOR}{content_column_value}"
            return content_column_value
    
    return concatenated_content


def validate_row_has_required_keys(
    values: List[str],
    required_keys_indices: Dict[int, str]
) -> bool:
    for index in required_keys_indices:
        if index >= len(values) or not values[index].strip():
            return False
    return True


def process_row(
    values: List[str],
    column_names: List[str],
    required_keys_indices: Dict[int, str],
    keys_outside_content_indices: Dict[int, str],
    content_column_index: int
) -> Optional[Dict[str, str]]:
    if not values or all(not v.strip() for v in values):
        return None
    
    while len(values) < len(column_names):
        values.append("")
    
    if not validate_row_has_required_keys(values, required_keys_indices):
        return None
    
    keys_outside_content_values: Dict[str, str] = {}
    for index, key in keys_outside_content_indices.items():
        if index < len(values):
            value = values[index].strip()
            if value:
                keys_outside_content_values[key] = value
    
    final_content = build_content(
        values,
        column_names,
        keys_outside_content_indices,
        content_column_index
    )
    
    return {
        COLUMN_NAME: keys_outside_content_values.get(COLUMN_NAME, ""),
        COLUMN_CONTENT: final_content,
        COLUMN_TYPE: keys_outside_content_values.get(COLUMN_TYPE, ""),
    }


def parse_csv(input_file_path: str) -> List[Dict[str, str]]:
    rows: List[Dict[str, str]] = []
    
    with open(input_file_path, 'r', encoding='utf-8', newline='') as csvfile:
        reader = csv.reader(csvfile)
        
        try:
            column_names = next(reader)
        except StopIteration:
            raise ValueError("No header found in CSV file")
        
        column_names = [col.strip() for col in column_names if col.strip()]
        
        if not column_names:
            raise ValueError("No column names found in CSV header")
        
        required_keys_indices = validate_required_keys(column_names)
        keys_outside_content_indices = find_keys_outside_content_indices(column_names)
        content_column_index = find_column_index(column_names, COLUMN_CONTENT)
        
        for row_values in reader:
            try:
                processed_row = process_row(
                    row_values,
                    column_names,
                    required_keys_indices,
                    keys_outside_content_indices,
                    content_column_index
                )
                if processed_row is not None:
                    rows.append(processed_row)
            except Exception as e:
                print(f"Failed to process CSV row: {e}")
                continue
    
    return rows


def escape_csv_value(value: str) -> str:
    if not value:
        return ""
    
    if ',' in value or '"' in value or '\n' in value:
        escaped_value = value.replace('"', '""')
        return f'"{escaped_value}"'
    
    return value


def generate_csv(rows: List[Dict[str, str]], output_file_path: str) -> None:
    headers = [COLUMN_NAME, COLUMN_CONTENT, COLUMN_TYPE]
    
    with open(output_file_path, 'w', encoding='utf-8', newline='') as csvfile:
        writer = csv.writer(csvfile)
        writer.writerow(headers)
        for row in rows:
            writer.writerow([
                row.get(COLUMN_NAME, ""),
                row.get(COLUMN_CONTENT, ""),
                row.get(COLUMN_TYPE, "")
            ])


def merge_csv_columns(input_file_path: str, output_file_path: str) -> None:
    rows = parse_csv(input_file_path)
    generate_csv(rows, output_file_path)


def generate_and_save_keys(keys_file_path: str = DEFAULT_KEYS_FILE) -> Dict[str, str]:
    if os.path.exists(keys_file_path):
        print(f"Keys file already exists at {keys_file_path}. Loading existing keys...")
        with open(keys_file_path, 'r', encoding='utf-8') as f:
            return json.load(f)
    
    print(f"Generating new encryption keys and saving to {keys_file_path}...")
    
    keys: Dict[str, str] = {}
    
    print("Generating RSA key pair for task inputs...")
    rsa_private_key, rsa_public_key = cryptography.generate_rsa_key_pair(key_size=4096)
    keys[KEY_NAMES["TASKS_INPUTS_RSA_PRIVATE_KEY"]] = rsa_private_key.decode('utf-8')
    keys[KEY_NAMES["TASKS_INPUTS_RSA_PUBLIC_KEY"]] = rsa_public_key.decode('utf-8')
    
    print("Generating ECDSA key pair for task inputs...")
    ecdsa_private_key, ecdsa_public_key = cryptography.generate_ecdsa_key_pair()
    keys[KEY_NAMES["TASKS_INPUTS_ECDSA_PRIVATE_KEY"]] = ecdsa_private_key.decode('utf-8')
    keys[KEY_NAMES["TASKS_INPUTS_ECDSA_PUBLIC_KEY"]] = ecdsa_public_key.decode('utf-8')
    
    print("Generating RSA key pair for task outputs...")
    rsa_private_key, rsa_public_key = cryptography.generate_rsa_key_pair(key_size=4096)
    keys[KEY_NAMES["TASKS_OUTPUTS_RSA_PRIVATE_KEY"]] = rsa_private_key.decode('utf-8')
    keys[KEY_NAMES["TASKS_OUTPUTS_RSA_PUBLIC_KEY"]] = rsa_public_key.decode('utf-8')
    
    print("Generating ECDSA key pair for task outputs...")
    ecdsa_private_key, ecdsa_public_key = cryptography.generate_ecdsa_key_pair()
    keys[KEY_NAMES["TASKS_OUTPUTS_ECDSA_PRIVATE_KEY"]] = ecdsa_private_key.decode('utf-8')
    keys[KEY_NAMES["TASKS_OUTPUTS_ECDSA_PUBLIC_KEY"]] = ecdsa_public_key.decode('utf-8')
    
    with open(keys_file_path, 'w', encoding='utf-8') as f:
        json.dump(keys, f, indent=2)
    
    print(f"Keys successfully saved to {keys_file_path}")
    return keys


def load_keys(keys_file_path: str = DEFAULT_KEYS_FILE) -> Dict[str, str]:
    if not os.path.exists(keys_file_path):
        raise FileNotFoundError(f"Keys file not found at {keys_file_path}. Run generate_and_save_keys() first.")
    
    with open(keys_file_path, 'r', encoding='utf-8') as f:
        return json.load(f)


def set_keys_in_settings(keys_file_path: str = DEFAULT_KEYS_FILE) -> None:
    from octobot_node.app.core.config import settings
    
    keys = load_keys(keys_file_path)
    
    def to_bytes(key_value: str) -> bytes:
        if isinstance(key_value, bytes):
            return key_value
        return key_value.encode('utf-8') if key_value else None
    
    settings.TASKS_INPUTS_RSA_PUBLIC_KEY = to_bytes(keys.get(KEY_NAMES["TASKS_INPUTS_RSA_PUBLIC_KEY"]))
    settings.TASKS_INPUTS_RSA_PRIVATE_KEY = to_bytes(keys.get(KEY_NAMES["TASKS_INPUTS_RSA_PRIVATE_KEY"]))
    settings.TASKS_INPUTS_ECDSA_PUBLIC_KEY = to_bytes(keys.get(KEY_NAMES["TASKS_INPUTS_ECDSA_PUBLIC_KEY"]))
    settings.TASKS_INPUTS_ECDSA_PRIVATE_KEY = to_bytes(keys.get(KEY_NAMES["TASKS_INPUTS_ECDSA_PRIVATE_KEY"]))
    
    settings.TASKS_OUTPUTS_RSA_PUBLIC_KEY = to_bytes(keys.get(KEY_NAMES["TASKS_OUTPUTS_RSA_PUBLIC_KEY"]))
    settings.TASKS_OUTPUTS_RSA_PRIVATE_KEY = to_bytes(keys.get(KEY_NAMES["TASKS_OUTPUTS_RSA_PRIVATE_KEY"]))
    settings.TASKS_OUTPUTS_ECDSA_PUBLIC_KEY = to_bytes(keys.get(KEY_NAMES["TASKS_OUTPUTS_ECDSA_PUBLIC_KEY"]))
    settings.TASKS_OUTPUTS_ECDSA_PRIVATE_KEY = to_bytes(keys.get(KEY_NAMES["TASKS_OUTPUTS_ECDSA_PRIVATE_KEY"]))
    
    print("Keys successfully loaded into settings")


def encrypt_csv_content(
    csv_rows: List[Dict[str, str]],
    content_column: str = "content"
) -> List[Dict[str, str]]:
    from octobot_node.app.core.config import settings
    
    if settings.TASKS_INPUTS_RSA_PUBLIC_KEY is None or settings.TASKS_INPUTS_ECDSA_PRIVATE_KEY is None:
        raise ValueError(
            f"Encryption keys are not set in settings. "
            f"TASKS_INPUTS_RSA_PUBLIC_KEY={settings.TASKS_INPUTS_RSA_PUBLIC_KEY is not None}, "
            f"TASKS_INPUTS_ECDSA_PRIVATE_KEY={settings.TASKS_INPUTS_ECDSA_PRIVATE_KEY is not None}. "
            f"Call set_keys_in_settings() or provide keys to merge_and_encrypt_csv() first."
        )
    
    encrypted_rows: List[Dict[str, str]] = []
    
    for row in csv_rows:
        encrypted_row = row.copy()
        content = row.get(content_column, "")
        
        if content:
            try:
                encrypted_content, metadata = encrypt_task_content(content)
                encrypted_row[content_column] = encrypted_content
                encrypted_row["metadata"] = metadata
            except Exception as e:
                error_msg = f"Failed to encrypt content for row '{row.get('name', 'unknown')}': {e}"
                raise Exception(error_msg) from e
        else:
            encrypted_row["metadata"] = ""
        
        encrypted_rows.append(encrypted_row)
    
    return encrypted_rows


def decrypt_csv_content(
    csv_rows: List[Dict[str, str]],
    content_column: str = "content",
    metadata_column: str = "metadata"
) -> List[Dict[str, str]]:
    decrypted_rows: List[Dict[str, str]] = []
    
    for row in csv_rows:
        decrypted_row = row.copy()
        encrypted_content = row.get(content_column, "")
        metadata = row.get(metadata_column, "")
        
        if encrypted_content and metadata:
            try:
                decrypted_content = decrypt_task_content(encrypted_content, metadata)
                decrypted_row[content_column] = decrypted_content
            except Exception as e:
                print(f"Failed to decrypt content for row '{row.get('name', 'unknown')}': {e}")
        elif not encrypted_content:
            pass
        else:
            print(f"Warning: Row '{row.get('name', 'unknown')}' has content but no metadata. Skipping decryption.")
        
        decrypted_row.pop(metadata_column, None)
        decrypted_rows.append(decrypted_row)
    
    return decrypted_rows


def encrypt_csv_file(
    input_file_path: str,
    output_file_path: str,
    content_column: str = "content"
) -> None:
    rows = parse_csv(input_file_path)
    encrypted_rows = encrypt_csv_content(rows, content_column)
    headers = ["name", content_column, "type", "metadata"]
    
    with open(output_file_path, 'w', encoding='utf-8', newline='') as csvfile:
        writer = csv.writer(csvfile)
        writer.writerow(headers)
        for row in encrypted_rows:
            writer.writerow([
                row.get("name", ""),
                row.get(content_column, ""),
                row.get("type", ""),
                row.get("metadata", "")
            ])


def decrypt_csv_file(
    input_file_path: str,
    output_file_path: str,
    content_column: str = "content",
    metadata_column: str = "metadata"
) -> None:
    rows = []
    with open(input_file_path, 'r', encoding='utf-8', newline='') as csvfile:
        reader = csv.DictReader(csvfile)
        for row in reader:
            rows.append(dict(row))
    decrypted_rows = decrypt_csv_content(rows, content_column, metadata_column)
    headers = ["name", content_column, "type"]
    
    with open(output_file_path, 'w', encoding='utf-8', newline='') as csvfile:
        writer = csv.writer(csvfile)
        writer.writerow(headers)
        
        for row in decrypted_rows:
            writer.writerow([
                row.get("name", ""),
                row.get(content_column, ""),
                row.get("type", "")
            ])


def merge_and_encrypt_csv(
    input_file_path: str,
    output_file_path: str,
    content_column: str = "content",
    keys: Optional[Dict[str, str]] = None,
    keys_file_path: Optional[str] = None
) -> None:
    if keys_file_path:
        set_keys_in_settings(keys_file_path)
    elif keys:
        from octobot_node.app.core.config import settings
        
        def to_bytes(key_value: str) -> bytes:
            if isinstance(key_value, bytes):
                return key_value
            return key_value.encode('utf-8') if key_value else None
        
        settings.TASKS_INPUTS_RSA_PUBLIC_KEY = to_bytes(keys.get(KEY_NAMES["TASKS_INPUTS_RSA_PUBLIC_KEY"]))
        settings.TASKS_INPUTS_RSA_PRIVATE_KEY = to_bytes(keys.get(KEY_NAMES["TASKS_INPUTS_RSA_PRIVATE_KEY"]))
        settings.TASKS_INPUTS_ECDSA_PUBLIC_KEY = to_bytes(keys.get(KEY_NAMES["TASKS_INPUTS_ECDSA_PUBLIC_KEY"]))
        settings.TASKS_INPUTS_ECDSA_PRIVATE_KEY = to_bytes(keys.get(KEY_NAMES["TASKS_INPUTS_ECDSA_PRIVATE_KEY"]))
        settings.TASKS_OUTPUTS_RSA_PUBLIC_KEY = to_bytes(keys.get(KEY_NAMES["TASKS_OUTPUTS_RSA_PUBLIC_KEY"]))
        settings.TASKS_OUTPUTS_RSA_PRIVATE_KEY = to_bytes(keys.get(KEY_NAMES["TASKS_OUTPUTS_RSA_PRIVATE_KEY"]))
        settings.TASKS_OUTPUTS_ECDSA_PUBLIC_KEY = to_bytes(keys.get(KEY_NAMES["TASKS_OUTPUTS_ECDSA_PUBLIC_KEY"]))
        settings.TASKS_OUTPUTS_ECDSA_PRIVATE_KEY = to_bytes(keys.get(KEY_NAMES["TASKS_OUTPUTS_ECDSA_PRIVATE_KEY"]))
    
    rows = parse_csv(input_file_path)
    encrypted_rows = encrypt_csv_content(rows, content_column)
    headers = ["name", content_column, "type", "metadata"]
    
    with open(output_file_path, 'w', encoding='utf-8', newline='') as csvfile:
        writer = csv.writer(csvfile)
        writer.writerow(headers)
        for row in encrypted_rows:
            writer.writerow([
                row.get("name", ""),
                row.get(content_column, ""),
                row.get("type", ""),
                row.get("metadata", "")
            ])



#  This file is part of OctoBot Node (https://github.com/Drakkar-Software/OctoBot-Node)
#  Copyright (c) 2025 Drakkar-Software, All rights reserved.
#
#  OctoBot is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  OctoBot is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#
#  You should have received a copy of the GNU General Public
#  License along with OctoBot. If not, see <https://www.gnu.org/licenses/>.

import uuid
from typing import Annotated

from fastapi import Depends, HTTPException, status
from fastapi.security import HTTPBasic, HTTPBasicCredentials

from octobot_node.app.core.config import settings
from octobot_node.app.models import User

security_basic = HTTPBasic()

_BASIC_AUTH_USER_ID = uuid.uuid4()

# TODO: support other auth methods (like supabase, jwt, etc.)

def get_current_user(credentials: Annotated[HTTPBasicCredentials, Depends(security_basic)]) -> User:
    if credentials.username != settings.ADMIN_USERNAME:
        raise HTTPException(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail="Incorrect username or password",
            headers={"WWW-Authenticate": "Basic"},
        )
    
    if credentials.password != settings.ADMIN_PASSWORD:
        raise HTTPException(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail="Incorrect username or password",
            headers={"WWW-Authenticate": "Basic"},
        )

    user = User(
        id=_BASIC_AUTH_USER_ID,
        email=settings.ADMIN_USERNAME,
        is_active=True,
        is_superuser=True,
        full_name=None,
    )
    return user


CurrentUser = Annotated[User, Depends(get_current_user)]


def get_current_active_superuser(current_user: CurrentUser) -> User:
    if not current_user.is_superuser:
        raise HTTPException(
            status_code=status.HTTP_403_FORBIDDEN,
            detail="The user doesn't have enough privileges"
        )
    return current_user



#  This file is part of OctoBot Node (https://github.com/Drakkar-Software/OctoBot-Node)
#  Copyright (c) 2025 Drakkar-Software, All rights reserved.
#
#  OctoBot is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  OctoBot is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#
#  You should have received a copy of the GNU General Public
#  License along with OctoBot. If not, see <https://www.gnu.org/licenses/>.

import uuid
import typing
import datetime
from enum import Enum

from pydantic import EmailStr
from pydantic import BaseModel, Field


class UserBase(BaseModel):
    email: EmailStr = Field(max_length=255)
    is_active: bool = True
    is_superuser: bool = False
    full_name: str | None = Field(default=None, max_length=255)


class User(UserBase):
    id: uuid.UUID


class TaskStatus(str, Enum):
    PENDING = "pending"
    SCHEDULED = "scheduled"
    PERIODIC = "periodic"
    RUNNING = "running"
    COMPLETED = "completed"
    FAILED = "failed"


class TaskType(str, Enum):
    START_OCTOBOT = "start_octobot"
    EXECUTE_ACTIONS = "execute_actions"
    STOP_OCTOBOT = "stop_octobot"

class Task(BaseModel):
    id: uuid.UUID = uuid.uuid4()
    name: typing.Optional[str] = None
    description: typing.Optional[str] = None
    content: typing.Optional[str] = None
    content_metadata: typing.Optional[str] = None
    type: typing.Optional[str] = None
    status: typing.Optional[TaskStatus] = None
    result: typing.Optional[str] = None
    result_metadata: typing.Optional[str] = None
    retries: typing.Optional[int] = 0
    retry_delay: typing.Optional[int] = 0
    priority: typing.Optional[int] = 0
    expires: typing.Optional[datetime.datetime] = None
    expires_resolved: typing.Optional[datetime.datetime] = None
    scheduled_at: typing.Optional[datetime.datetime] = None
    started_at: typing.Optional[datetime.datetime] = None
    completed_at: typing.Optional[datetime.datetime] = None

class Node(BaseModel):
    node_type: str
    backend_type: str
    workers: int | None
    status: str
    redis_url: str | None = None
    sqlite_file: str | None = None



#  This file is part of OctoBot Node (https://github.com/Drakkar-Software/OctoBot-Node)
#  Copyright (c) 2025 Drakkar-Software, All rights reserved.
#
#  OctoBot is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  OctoBot is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#
#  You should have received a copy of the GNU General Public
#  License along with OctoBot. If not, see <https://www.gnu.org/licenses/>.

from datetime import datetime, timedelta, timezone
from typing import Any

import jwt
from passlib.context import CryptContext

from octobot_node.app.core.config import settings

pwd_context = CryptContext(schemes=["bcrypt"], deprecated="auto")


ALGORITHM = "HS256"


def create_access_token(subject: str | Any, expires_delta: timedelta) -> str:
    expire = datetime.now(timezone.utc) + expires_delta
    to_encode = {"exp": expire, "sub": str(subject)}
    encoded_jwt = jwt.encode(to_encode, settings.SECRET_KEY, algorithm=ALGORITHM)
    return encoded_jwt


def verify_password(plain_password: str, hashed_password: str) -> bool:
    return pwd_context.verify(plain_password, hashed_password)


def get_password_hash(password: str) -> str:
    return pwd_context.hash(password)



#  This file is part of OctoBot Node (https://github.com/Drakkar-Software/OctoBot-Node)
#  Copyright (c) 2025 Drakkar-Software, All rights reserved.
#
#  OctoBot is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  OctoBot is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#
#  You should have received a copy of the GNU General Public
#  License along with OctoBot. If not, see <https://www.gnu.org/licenses/>.

from typing import Any, List, Tuple
import uuid
from fastapi import APIRouter

from octobot_node.app.models import Task
from octobot_node.scheduler.api import get_all_tasks, get_task_metrics
from octobot_node.scheduler.tasks import trigger_task

router = APIRouter(tags=["tasks"])

@router.post("/", response_model=Tuple[int, int])
def create_tasks(tasks: List[Task]) -> Tuple[int, int]:
    success_count = 0
    error_count = 0
    for task in tasks:
        is_scheduled = trigger_task(task)
        if is_scheduled:
            success_count += 1
        else:
            error_count += 1
    return success_count, error_count


@router.get("/metrics")
def get_metrics() -> Any:
    return get_task_metrics()

@router.get("/", response_model=List[Task])
def get_tasks(page: int = 1, limit: int = 100) -> Any:
    tasks_data = get_all_tasks()
    
    start_idx = (page - 1) * limit
    end_idx = start_idx + limit
    paginated_tasks = tasks_data[start_idx:end_idx]
    return paginated_tasks

@router.put("/", response_model=Task)
def update_task(taskId: uuid.UUID, task: Task) -> Any:
    # TODO
    return task

@router.delete("/", response_model=str)
def delete_task(taskId: uuid.UUID) -> str:
    # TODO
    return taskId



#  Drakkar-Software OctoBot-Node
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

import pytest
from unittest.mock import MagicMock
from typing import Any

from pytest_mock import MockerFixture

from octobot_node.scheduler.api import (
    get_node_status,
    get_task_metrics,
    get_all_tasks,
    get_task_result,
)


class TestGetNodeStatus:
    """Tests for get_node_status function."""

    def test_get_node_status_master_node_with_redis(
        self, mocker: MockerFixture
    ) -> None:
        """Test node status for master node with Redis backend."""
        mock_settings = mocker.patch("octobot_node.scheduler.api.settings")
        mock_settings.IS_MASTER_MODE = True
        mock_settings.SCHEDULER_REDIS_URL = "redis://localhost:6379"
        mock_settings.SCHEDULER_SQLITE_FILE = "tasks.db"
        mock_settings.SCHEDULER_WORKERS = 0

        mock_consumer = mocker.patch("octobot_node.scheduler.api.CONSUMER")
        mock_consumer.is_started.return_value = False

        result = get_node_status()

        assert result["node_type"] == "master"
        assert result["backend_type"] == "redis"
        assert result["workers"] is None
        assert result["status"] == "running"
        assert result["redis_url"] == "redis://localhost:6379"
        assert result["sqlite_file"] is None

    def test_get_node_status_slave_node_with_sqlite(
        self, mocker: MockerFixture
    ) -> None:
        """Test node status for consumer node with SQLite backend."""
        mock_settings = mocker.patch("octobot_node.scheduler.api.settings")
        mock_settings.IS_MASTER_MODE = False
        mock_settings.SCHEDULER_REDIS_URL = None
        mock_settings.SCHEDULER_SQLITE_FILE = "tasks.db"
        mock_settings.SCHEDULER_WORKERS = 4

        mock_consumer = mocker.patch("octobot_node.scheduler.api.CONSUMER")
        mock_consumer.is_started.return_value = True
        mock_consumer.workers = 4

        result = get_node_status()

        assert result["node_type"] == "consumer"
        assert result["backend_type"] == "sqlite"
        assert result["workers"] == 4
        assert result["status"] == "running"
        assert result["redis_url"] is None
        assert result["sqlite_file"] == "tasks.db"

    def test_get_node_status_slave_node_stopped(
        self, mocker: MockerFixture
    ) -> None:
        """Test node status for consumer node when consumer is stopped."""
        mock_settings = mocker.patch("octobot_node.scheduler.api.settings")
        mock_settings.IS_MASTER_MODE = False
        mock_settings.SCHEDULER_REDIS_URL = None
        mock_settings.SCHEDULER_SQLITE_FILE = "tasks.db"
        mock_settings.SCHEDULER_WORKERS = 4

        mock_consumer = mocker.patch("octobot_node.scheduler.api.CONSUMER")
        mock_consumer.is_started.return_value = False
        mock_consumer.workers = 4

        result = get_node_status()

        assert result["node_type"] == "consumer"
        assert result["status"] == "stopped"
        assert result["workers"] == 4

    def test_get_node_status_master_node_always_running(
        self, mocker: MockerFixture
    ) -> None:
        """Test that master node is always running regardless of consumer state."""
        mock_settings = mocker.patch("octobot_node.scheduler.api.settings")
        mock_settings.IS_MASTER_MODE = True
        mock_settings.SCHEDULER_REDIS_URL = None
        mock_settings.SCHEDULER_SQLITE_FILE = "tasks.db"
        mock_settings.SCHEDULER_WORKERS = 0

        mock_consumer = mocker.patch("octobot_node.scheduler.api.CONSUMER")
        mock_consumer.is_started.return_value = False

        result = get_node_status()

        assert result["status"] == "running"
        assert result["node_type"] == "master"

    def test_get_node_status_both_master_and_consumers(
        self, mocker: MockerFixture
    ) -> None:
        """Test node status when both master mode and consumers are enabled."""
        mock_settings = mocker.patch("octobot_node.scheduler.api.settings")
        mock_settings.IS_MASTER_MODE = True
        mock_settings.SCHEDULER_REDIS_URL = "redis://localhost:6379"
        mock_settings.SCHEDULER_SQLITE_FILE = "tasks.db"
        mock_settings.SCHEDULER_WORKERS = 4

        mock_consumer = mocker.patch("octobot_node.scheduler.api.CONSUMER")
        mock_consumer.is_started.return_value = True
        mock_consumer.workers = 4

        result = get_node_status()

        assert result["node_type"] == "both"
        assert result["backend_type"] == "redis"
        assert result["workers"] == 4
        assert result["status"] == "running"

    def test_get_node_status_none(
        self, mocker: MockerFixture
    ) -> None:
        """Test node status when neither master mode nor consumers are enabled."""
        mock_settings = mocker.patch("octobot_node.scheduler.api.settings")
        mock_settings.IS_MASTER_MODE = False
        mock_settings.SCHEDULER_REDIS_URL = None
        mock_settings.SCHEDULER_SQLITE_FILE = "tasks.db"
        mock_settings.SCHEDULER_WORKERS = 0

        mock_consumer = mocker.patch("octobot_node.scheduler.api.CONSUMER")
        mock_consumer.is_started.return_value = False

        result = get_node_status()

        assert result["node_type"] == "none"
        assert result["status"] == "stopped"
        assert result["workers"] is None


class TestGetTaskMetrics:
    """Tests for get_task_metrics function."""

    def test_get_task_metrics_success(
        self, mocker: MockerFixture
    ) -> None:
        """Test successful retrieval of task metrics."""
        mock_huey = MagicMock()
        mock_huey.pending_count.return_value = 5
        mock_huey.scheduled_count.return_value = 3
        mock_huey.result_count.return_value = 10

        mock_scheduler = mocker.patch("octobot_node.scheduler.api.SCHEDULER")
        mock_scheduler.INSTANCE = mock_huey
        mock_scheduler.get_periodic_tasks.return_value = [
            {"id": "task1"},
            {"id": "task2"},
        ]

        result = get_task_metrics()

        assert result["pending"] == 5
        assert result["scheduled"] == 5  # 3 + 2 periodic tasks
        assert result["results"] == 10
        mock_huey.pending_count.assert_called_once()
        mock_huey.scheduled_count.assert_called_once()
        mock_huey.result_count.assert_called_once()
        mock_scheduler.get_periodic_tasks.assert_called_once()

    def test_get_task_metrics_uninitialized_scheduler(
        self, mocker: MockerFixture
    ) -> None:
        """Test task metrics when scheduler is not initialized."""
        mock_scheduler = mocker.patch("octobot_node.scheduler.api.SCHEDULER")
        mock_scheduler.INSTANCE = None

        result = get_task_metrics()

        assert result == {"pending": 0, "scheduled": 0, "results": 0}

    def test_get_task_metrics_exception_handling(
        self, mocker: MockerFixture
    ) -> None:
        """Test task metrics when an exception occurs."""
        mock_scheduler = mocker.patch("octobot_node.scheduler.api.SCHEDULER")
        mock_scheduler.INSTANCE = MagicMock()
        mock_scheduler.INSTANCE.pending_count.side_effect = Exception("Database error")

        result = get_task_metrics()

        assert result == {"pending": 0, "scheduled": 0, "results": 0}

    def test_get_task_metrics_no_periodic_tasks(
        self, mocker: MockerFixture
    ) -> None:
        """Test task metrics when there are no periodic tasks."""
        mock_huey = MagicMock()
        mock_huey.pending_count.return_value = 2
        mock_huey.scheduled_count.return_value = 1
        mock_huey.result_count.return_value = 5

        mock_scheduler = mocker.patch("octobot_node.scheduler.api.SCHEDULER")
        mock_scheduler.INSTANCE = mock_huey
        mock_scheduler.get_periodic_tasks.return_value = []

        result = get_task_metrics()

        assert result["pending"] == 2
        assert result["scheduled"] == 1
        assert result["results"] == 5


class TestGetAllTasks:
    """Tests for get_all_tasks function."""

    def test_get_all_tasks_success(
        self, mocker: MockerFixture
    ) -> None:
        """Test successful retrieval of all tasks."""
        periodic_tasks = [{"id": "periodic1", "status": "periodic"}]
        pending_tasks = [{"id": "pending1", "status": "pending"}]
        scheduled_tasks = [{"id": "scheduled1", "status": "scheduled"}]
        result_tasks = [{"id": "result1", "status": "completed"}]

        mock_scheduler = mocker.patch("octobot_node.scheduler.api.SCHEDULER")
        mock_scheduler.get_periodic_tasks.return_value = periodic_tasks
        mock_scheduler.get_pending_tasks.return_value = pending_tasks
        mock_scheduler.get_scheduled_tasks.return_value = scheduled_tasks
        mock_scheduler.get_results.return_value = result_tasks

        result = get_all_tasks()

        assert len(result) == 4
        assert periodic_tasks[0] in result
        assert pending_tasks[0] in result
        assert scheduled_tasks[0] in result
        assert result_tasks[0] in result
        mock_scheduler.get_periodic_tasks.assert_called_once()
        mock_scheduler.get_pending_tasks.assert_called_once()
        mock_scheduler.get_scheduled_tasks.assert_called_once()
        mock_scheduler.get_results.assert_called_once()

    def test_get_all_tasks_empty(
        self, mocker: MockerFixture
    ) -> None:
        """Test get_all_tasks when there are no tasks."""
        mock_scheduler = mocker.patch("octobot_node.scheduler.api.SCHEDULER")
        mock_scheduler.get_periodic_tasks.return_value = []
        mock_scheduler.get_pending_tasks.return_value = []
        mock_scheduler.get_scheduled_tasks.return_value = []
        mock_scheduler.get_results.return_value = []

        result = get_all_tasks()

        assert result == []

    def test_get_all_tasks_exception_handling(
        self, mocker: MockerFixture
    ) -> None:
        """Test get_all_tasks when an exception occurs."""
        mock_scheduler = mocker.patch("octobot_node.scheduler.api.SCHEDULER")
        mock_scheduler.get_periodic_tasks.side_effect = Exception("Database error")

        result = get_all_tasks()

        assert result == []

    def test_get_all_tasks_partial_exception(
        self, mocker: MockerFixture
    ) -> None:
        """Test get_all_tasks when some methods fail but others succeed."""
        periodic_tasks = [{"id": "periodic1"}]
        pending_tasks = [{"id": "pending1"}]

        mock_scheduler = mocker.patch("octobot_node.scheduler.api.SCHEDULER")
        mock_scheduler.get_periodic_tasks.return_value = periodic_tasks
        mock_scheduler.get_pending_tasks.return_value = pending_tasks
        mock_scheduler.get_scheduled_tasks.side_effect = Exception("Error")
        mock_scheduler.get_results.return_value = []

        result = get_all_tasks()

        assert len(result) == 2
        assert periodic_tasks[0] in result
        assert pending_tasks[0] in result


class TestGetTaskResult:
    """Tests for get_task_result function."""

    @pytest.mark.asyncio
    async def test_get_task_result_completed(
        self, mocker: MockerFixture
    ) -> None:
        """Test get_task_result for a completed task."""
        task_id = "task-123"
        result_data = {"status": "success", "output": "completed"}

        mock_result = MagicMock()
        mock_result.ready.return_value = True
        mock_result.get.return_value = result_data

        mock_scheduler = mocker.patch("octobot_node.scheduler.api.SCHEDULER")
        mock_scheduler.INSTANCE = MagicMock()
        mock_scheduler.INSTANCE.result.return_value = mock_result

        result = await get_task_result(task_id)

        assert result["status"] == "completed"
        assert result["data"] == result_data
        mock_scheduler.INSTANCE.result.assert_called_once_with(task_id)
        mock_result.ready.assert_called_once()
        mock_result.get.assert_called_once()

    @pytest.mark.asyncio
    async def test_get_task_result_pending(
        self, mocker: MockerFixture
    ) -> None:
        """Test get_task_result for a pending task."""
        task_id = "task-456"

        mock_result = MagicMock()
        mock_result.ready.return_value = False

        mock_scheduler = mocker.patch("octobot_node.scheduler.api.SCHEDULER")
        mock_scheduler.INSTANCE = MagicMock()
        mock_scheduler.INSTANCE.result.return_value = mock_result

        result = await get_task_result(task_id)

        assert result["status"] == "pending or running"
        assert "data" not in result
        mock_scheduler.INSTANCE.result.assert_called_once_with(task_id)
        mock_result.ready.assert_called_once()
        mock_result.get.assert_not_called()

    @pytest.mark.asyncio
    async def test_get_task_result_not_found(
        self, mocker: MockerFixture
    ) -> None:
        """Test get_task_result for a task that doesn't exist."""
        task_id = "task-789"

        mock_scheduler = mocker.patch("octobot_node.scheduler.api.SCHEDULER")
        mock_scheduler.INSTANCE = MagicMock()
        mock_scheduler.INSTANCE.result.return_value = None

        result = await get_task_result(task_id)

        assert result == {"error": "task not found"}
        mock_scheduler.INSTANCE.result.assert_called_once_with(task_id)

    @pytest.mark.asyncio
    async def test_get_task_result_running(
        self, mocker: MockerFixture
    ) -> None:
        """Test get_task_result for a running task."""
        task_id = "task-running"

        mock_result = MagicMock()
        mock_result.ready.return_value = False

        mock_scheduler = mocker.patch("octobot_node.scheduler.api.SCHEDULER")
        mock_scheduler.INSTANCE = MagicMock()
        mock_scheduler.INSTANCE.result.return_value = mock_result

        result = await get_task_result(task_id)

        assert result["status"] == "pending or running"


#  Drakkar-Software OctoBot-Node
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

from pathlib import Path

import pytest

from tests.scheduler.task_import.csv_utils import (
    generate_and_save_keys,
    merge_and_encrypt_csv,
    decrypt_csv_file,
    parse_csv,
    merge_csv_columns,
    DEFAULT_KEYS_FILE,
)


class TestCSVEncryption:
    def test_encrypt_and_decrypt_csv(self, tmp_path: Path) -> None:
        test_dir = Path(__file__).parent
        test_csv_path = test_dir / "test-tasks.csv"
        keys_file = str(test_dir / DEFAULT_KEYS_FILE)
        
        encrypted_csv = test_dir / "encrypted_tasks.csv"
        decrypted_csv = tmp_path / "decrypted_tasks.csv"
        merged_csv = tmp_path / "merged_tasks.csv"
        
        generate_and_save_keys(keys_file)
        
        merge_csv_columns(str(test_csv_path), str(merged_csv))
        
        from octobot_node.app.core.config import settings
        from tests.scheduler.task_import.csv_utils import set_keys_in_settings
        set_keys_in_settings(keys_file)
        
        assert settings.TASKS_INPUTS_RSA_PUBLIC_KEY is not None, "RSA public key should be set"
        assert settings.TASKS_INPUTS_ECDSA_PRIVATE_KEY is not None, "ECDSA private key should be set"
        
        merge_and_encrypt_csv(
            str(test_csv_path),
            str(encrypted_csv),
            keys_file_path=str(keys_file)
        )
        
        assert encrypted_csv.exists(), "Encrypted CSV file should be created"
        
        decrypt_csv_file(
            str(encrypted_csv),
            str(decrypted_csv)
        )
        
        assert decrypted_csv.exists(), "Decrypted CSV file should be created"
        
        original_rows = parse_csv(str(merged_csv))
        decrypted_rows = parse_csv(str(decrypted_csv))
        
        assert len(original_rows) == len(decrypted_rows), "Number of rows should match"
        
        for original_row, decrypted_row in zip(original_rows, decrypted_rows):
            assert original_row["name"] == decrypted_row["name"], "Names should match"
            assert original_row["type"] == decrypted_row["type"], "Types should match"
            assert original_row["content"] == decrypted_row["content"], "Content should match after decryption"



#  This file is part of OctoBot Node (https://github.com/Drakkar-Software/OctoBot-Node)
#  Copyright (c) 2025 Drakkar-Software, All rights reserved.
#
#  OctoBot is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  OctoBot is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#
#  You should have received a copy of the GNU General Public
#  License along with OctoBot. If not, see <https://www.gnu.org/licenses/>.
import os
import subprocess
import sys
from setuptools import find_packages
from setuptools import setup
from setuptools.command.sdist import sdist
from octobot_node import PROJECT_NAME, AUTHOR, VERSION

PACKAGES = find_packages(exclude=["tests", ])

# long description from README file
with open('README.md', encoding='utf-8') as f:
    DESCRIPTION = f.read()


REQUIRED = open('requirements.txt').readlines()
REQUIRES_PYTHON = '>=3.10'


class BuildUIAndSDist(sdist):    
    def run(self):
        self.announce('Running npm build before creating source distribution...', level=2)
        try:
            self.announce('Installing npm dependencies...', level=2)
            subprocess.check_call(['npm', 'install'], cwd=os.getcwd())
            subprocess.check_call(['npm', 'run', 'build'], cwd=os.getcwd())
            self.announce('npm build completed successfully', level=2)
        except subprocess.CalledProcessError as e:
            self.announce(f'Error running npm build: {e}', level=1)
            sys.exit(1)
        except FileNotFoundError:
            self.announce('npm not found. Skipping npm build step.', level=1)
                
        # Call the parent sdist run method
        super().run()


setup(
    name=PROJECT_NAME.lower().replace("-", "_"),
    version=VERSION,
    url='https://github.com/Drakkar-Software/OctoBot-Node',
    license='GPL-3.0',
    author=AUTHOR,
    author_email='contact@drakkar.software',
    description='OctoBot Node',
    py_modules=['start'],
    packages=PACKAGES,
    package_data={
        "": ["config/*"],
        "octobot_node.ui": ["dist/**/*"],
    },
    include_package_data=True,
    long_description=DESCRIPTION,
    long_description_content_type='text/markdown',
    tests_require=["pytest"],
    test_suite="tests",
    zip_safe=False,
    install_requires=REQUIRED,
    python_requires=REQUIRES_PYTHON,
    entry_points={
        'console_scripts': [
            PROJECT_NAME.replace("-", "_") + ' = octobot_node.cli:main'
        ]
    },
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Operating System :: OS Independent',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: POSIX',
        'Programming Language :: Python :: 3.10',
    ],
    cmdclass={
        'sdist': BuildUIAndSDist,
    },
)



#  This file is part of OctoBot Node (https://github.com/Drakkar-Software/OctoBot-Node)
#  Copyright (c) 2025 Drakkar-Software, All rights reserved.
#
#  OctoBot is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  OctoBot is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#
#  You should have received a copy of the GNU General Public
#  License along with OctoBot. If not, see <https://www.gnu.org/licenses/>.

from pathlib import Path


def get_dist_directory() -> Path | None:
    """
    Get the path to the dist directory containing the built frontend assets.
    
    The dist directory is always located at octobot_node/ui/dist,
    both in development and when installed as a package.
    
    Returns:
        Path to dist directory if found, None otherwise.
    """
    # Locate octobot_node.ui module directory
    try:
        import octobot_node.ui
        ui_module_path = Path(octobot_node.ui.__file__).resolve().parent
        dist_path = ui_module_path / "dist"
        if dist_path.exists() and dist_path.is_dir():
            return dist_path
    except (ImportError, ModuleNotFoundError, AttributeError):
        pass
    
    # Fallback: try relative to current file (for development if module not found)
    # Go up from octobot_node/app/utils.py -> octobot_node/app -> octobot_node -> ui -> dist
    current_file = Path(__file__).resolve()
    ui_module_path = current_file.parent.parent / "ui"
    dist_path = ui_module_path / "dist"
    
    if dist_path.exists() and dist_path.is_dir():
        return dist_path
    
    return None




#  This file is part of OctoBot Node (https://github.com/Drakkar-Software/OctoBot-Node)
#  Copyright (c) 2025 Drakkar-Software, All rights reserved.
#
#  OctoBot is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  OctoBot is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#
#  You should have received a copy of the GNU General Public
#  License along with OctoBot. If not, see <https://www.gnu.org/licenses/>.

import logging
import secrets
import sys
from typing import Annotated, Any, Literal

from pydantic import (
    AnyUrl,
    BeforeValidator,
    EmailStr,
    HttpUrl,
    computed_field,
    model_validator,
)
from pydantic_settings import BaseSettings, SettingsConfigDict
from typing_extensions import Self


DEFAULT_ADMIN_USERNAME: EmailStr = "admin@example.com"
DEFAULT_ADMIN_PASSWORD: str = "changethis"

def _get_env_file() -> str:
    # Check if pytest module is imported
    if "pytest" in sys.modules:
        return ".env.test"
    return ".env"


def parse_cors(v: Any) -> list[str] | str:
    if isinstance(v, str) and not v.startswith("["):
        return [i.strip() for i in v.split(",") if i.strip()]
    elif isinstance(v, list | str):
        return v
    raise ValueError(v)


def parse_key_to_bytes(v: str | bytes | None) -> bytes | None:
    if v is None:
        return None
    if isinstance(v, bytes):
        return v
    return v.encode('utf-8')


class Settings(BaseSettings):
    model_config = SettingsConfigDict(
        # Use .env.test when running tests, otherwise use .env
        env_file=_get_env_file(),
        env_ignore_empty=True,
        extra="ignore",
    )
    API_V1_STR: str = "/api/v1"
    SECRET_KEY: str = secrets.token_urlsafe(32)
    # 60 minutes * 24 hours * 8 days = 8 days
    ACCESS_TOKEN_EXPIRE_MINUTES: int = 60 * 24 * 8
    ENVIRONMENT: Literal["local", "production"] = "production"
    BACKEND_HOST: str = "http://localhost:8000"
    FRONTEND_HOST: str = "http://localhost:5173" if ENVIRONMENT == "local" else BACKEND_HOST

    BACKEND_CORS_ORIGINS: Annotated[
        list[AnyUrl] | str, BeforeValidator(parse_cors)
    ] = []

    @computed_field  # type: ignore[prop-decorator]
    @property
    def all_cors_origins(self) -> list[str]:
        return [str(origin).rstrip("/") for origin in self.BACKEND_CORS_ORIGINS] + [
            self.FRONTEND_HOST
        ]

    SENTRY_DSN: HttpUrl | None = None
    SCHEDULER_REDIS_URL: AnyUrl | None = None
    SCHEDULER_SQLITE_FILE: str = "tasks.db"
    SCHEDULER_WORKERS: int = 0  # 0 disables consumers, >0 enables consumers
    IS_MASTER_MODE: bool = False  # Enable master node mode
    REDIS_STORAGE_CERTS_PATH: str | None = None

    ADMIN_USERNAME: EmailStr = DEFAULT_ADMIN_USERNAME
    ADMIN_PASSWORD: str = DEFAULT_ADMIN_PASSWORD

    # Used to decrypt inputs and encrypt outputs
    TASKS_INPUTS_RSA_PRIVATE_KEY: Annotated[bytes | None, BeforeValidator(parse_key_to_bytes)] = None
    TASKS_INPUTS_ECDSA_PUBLIC_KEY: Annotated[bytes | None, BeforeValidator(parse_key_to_bytes)] = None
    TASKS_OUTPUTS_RSA_PUBLIC_KEY: Annotated[bytes | None, BeforeValidator(parse_key_to_bytes)] = None
    TASKS_OUTPUTS_ECDSA_PRIVATE_KEY: Annotated[bytes | None, BeforeValidator(parse_key_to_bytes)] = None

    # Used to encrypt inputs and decrypt outputs
    TASKS_INPUTS_RSA_PUBLIC_KEY: Annotated[bytes | None, BeforeValidator(parse_key_to_bytes)] = None
    TASKS_INPUTS_ECDSA_PRIVATE_KEY: Annotated[bytes | None, BeforeValidator(parse_key_to_bytes)] = None
    TASKS_OUTPUTS_RSA_PRIVATE_KEY: Annotated[bytes | None, BeforeValidator(parse_key_to_bytes)] = None
    TASKS_OUTPUTS_ECDSA_PUBLIC_KEY: Annotated[bytes | None, BeforeValidator(parse_key_to_bytes)] = None

    def _check_default_secret(self, var_name: str, value: str | None, default_value: EmailStr | None) -> None:
        if value == default_value:
            message = (
                f'The value of {var_name} is "{default_value}", '
                "for security, please change it, at least for deployments."
            )
            if self.ENVIRONMENT == "local":
                logging.getLogger("Settings").warning(message)
            else:
                raise ValueError(message)

    @model_validator(mode="after")
    def _enforce_non_default_secrets(self) -> Self:
        if self.IS_MASTER_MODE:
            self._check_default_secret("ADMIN_USERNAME", self.ADMIN_USERNAME, DEFAULT_ADMIN_USERNAME)
            self._check_default_secret(
                "ADMIN_PASSWORD", self.ADMIN_PASSWORD, DEFAULT_ADMIN_PASSWORD
            )
        return self


settings = Settings()  # type: ignore